# -*- coding: utf-8 -*-
"""Models of evolution of codons.

Copyright 2011, Laurent Guéguen, <laurent.gueguen@univ-lyon1.fr>.
"""
__author__ = "Laurent Guéguen <laurent.gueguen@univ-lyon1.fr>"
__date__ = "2011 May 3"
__credits__ = """Guido van Rossum, for an excellent programming language."""

from nuclmodel import NuclModel
from model import Model 

stdcodgen={"TTT":"F", "TTC":"F", "TTA":"L", "TTG":"L", "CTT":"L", "CTC":"L", "CTA":"L", "CTG":"L", "ATT":"I", "ATC":"I", "ATA":"I", "ATG":"M", "GTT":"V", "GTC":"V", "GTA":"V", "GTG":"V", "TGT":"C", "TGC":"C", "TGA":"*", "TGG":"W", "CGT":"R", "CGC":"R", "CGA":"R", "CGG":"R", "AGT":"S", "AGC":"S", "AGA":"R", "AGG":"R", "GGT":"G", "GGC":"G", "GGA":"G", "GGG":"G", "TAT":"Y", "TAC":"Y", "TAA":"*" , "TAG":"*", "CAT":"H", "CAC":"H", "CAA":"Q", "CAG":"Q", "AAT":"N", "AAC":"N", "AAA":"K", "AAG":"K", "GAT":"D", "GAC":"D", "GAA":"E", "GAG":"E", "TCT":"S", "TCC":"S", "TCA":"S", "TCG":"S", "CCT":"P", "CCC":"P", "CCA":"P", "CCG":"P", "ACT":"T", "ACC":"T", "ACA":"T", "ACG":"T", "GCT":"A", "GCC":"A", "GCA":"A", "GCG":"A"}


class CodonModel:
    """ Class of Models for codons defined as a dictionnary {phase:NuclModel}.

    A CodonModel is built from a unique NuclModel, with additional omega parameter, defined as
    a multiplicative factor on non-synonymous substitutions (default: 1). The genetic code is the standard one.

    The syntax to build a model is like:
    
    m=CodonModel(model="T92(theta=0.1,rCgT=10)",omega=0.01)
"""
    
    def __init__(self,**kw):
        self.__dModel={}
        
        if kw.has_key("nuclmodel"):
            m=NuclModel(model=kw["model"])
        else:
            m=Model()
        
        omega=kw.get("omega",1)

        for i in range(3):
            self.__dModel[i]=Model()

        model=self.__dModel[0]
        for x in "ACGT":
            lx=m.next_extended(x)
            for u in "ACGT":
                for v in "ACGT":  #codon xuv
                    uv=u+v
                    for nx in lx:
                        if uv.startswith(nx[1][:2]):
                            f=nx[3]
                            if stdcodgen[x+uv]!=stdcodgen[nx[2]+uv]:
                                f*=omega
                            if len(uv)< len(nx[1]):
                                po=nx[1].lower()
                            else:
                                po=uv.lower()
                            if not ((stdcodgen[x+uv]=='*') ^ (stdcodgen[nx[2]+uv]=="*")):
                                model._Proportion__add(nx[0].lower()+x+po,nx[2],f)
        model.update()
        
        model=self.__dModel[1]
        for x in "ACGT":
            lx=m.next_extended(x)
            for u in "ACGT":
                for v in "ACGT":  #codon uxv
                    for nx in lx:
                        if u.startswith(nx[0][-1:]) and v.startswith(nx[1][:1]):
                            f=nx[3]
                            if stdcodgen[u+x+v]!=stdcodgen[u+nx[2]+v]:
                                f*=omega
                            if len(u)< len(nx[0]):
                                pr=nx[0].lower()
                            else:
                                pr=u.lower()
                            if len(v)< len(nx[1]):
                                po=nx[1].lower()
                            else:
                                po=v.lower()
                            if not ((stdcodgen[u+x+v]=='*') ^ (stdcodgen[u+nx[2]+v]=="*")):
                                model._Proportion__add(pr+x+po,nx[2],f)
        model.update()
        
        model=self.__dModel[2]
        for x in "ACGT":
            lx=m.next_extended(x)
            for u in "ACGT":
                for v in "ACGT":  #codon uvx
                    for nx in lx:
                        uv=u+v
                        if uv.startswith(nx[0][-2:]):
                            f=nx[3]
                            if stdcodgen[uv+x]!=stdcodgen[uv+nx[2]]:
                                f*=omega
                            if len(uv)< len(nx[0]):
                                pr=nx[0].lower()
                            else:
                                pr=uv.lower()
                            if not ((stdcodgen[uv+x]=='*') ^ (stdcodgen[uv+nx[2]]=="*")):
                                model._Proportion__add(pr+x+nx[1].lower(),nx[2],f)
    
        model.update()
        
        self.__max=max(map(lambda x: x.get_max(),self.__dModel.values()))
        
    def __getitem__(self,i):
        return self.__dModel[i]

    def get_max(self):
        return self.__max
    
    def next_all(self, pos):
        return self.__dModel[pos % 3].next_all()
