# -*- coding: utf-8 -*-
"""Usual models of evolution of nucleotide sequences.

Copyright 2009, Laurent Guéguen, <laurent.gueguen@univ-lyon1.fr>.
"""

__author__ = "Laurent Guéguen <laurent.gueguen@univ-lyon1.fr>"
__date__ = "2009 September 10"
__credits__ = """Bio++ site.  http://biopp.univ-montp2.fr/ """

from model import Model


class BibliNucl:

    @staticmethod
    def JC69(**kw):
        """ Jukes and Cantor 1969.
        Normalised such that the substitution rate on a stationnary
        sequence is 1.
        
        eta: rate (default=1)
        """

        eta=kw.get("eta",1)

        for s in ["eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        eta2=eta/3.0
    
        s=""
        s+="A|C %f\n"%eta2
        s+="A|G %f\n"%eta2
        s+="A|T %f\n"%eta2
        s+="C|A %f\n"%eta2
        s+="C|G %f\n"%eta2
        s+="C|T %f\n"%eta2
        s+="G|A %f\n"%eta2
        s+="G|C %f\n"%eta2
        s+="G|T %f\n"%eta2
        s+="T|A %f\n"%eta2
        s+="T|C %f\n"%eta2
        s+="T|G %f\n"%eta2
        s+="Cg|T %f\n"%(rCT*eta2)
        s+="cG|A %f\n"%(rCT*eta2)

        m=model.Model(str=s)
        return m

    @staticmethod
    def K80(**kw):
        """ Kimura 1980.
        Normalized such that the substitution rate on a stationnary
        sequence is 1.
        
        kappa: transition/transversion rate.
        eta: rate (default=1)
        """
        kappa=kw.get("kappa",1)
        rCgT=kw.get("rCgT",0)
        rcGA=kw.get("rcGA",0)
        eta=kw.get("eta",1)
        
        for s in ["kappa","rCgT","rcGA","eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)
            
            eta2=eta/(kappa+2.0)
            
        s=""
        s+="A|C %f\n"%eta2
        s+="A|G %f\n"%(kappa*eta2)
        s+="A|T %f\n"%eta2
        s+="C|A %f\n"%eta2
        s+="C|G %f\n"%eta2
        s+="C|T %f\n"%(kappa*eta2)
        s+="G|A %f\n"%(kappa*eta2)
        s+="G|C %f\n"%eta2
        s+="G|T %f\n"%eta2
        s+="T|A %f\n"%eta2
        s+="T|C %f\n"%(kappa*eta2)
        s+="T|G %f\n"%eta2
        
        return s
    
    @staticmethod
    def RN95(**kw):
        """ Model first introduced in Rhetzky & Ney, 1995 and
        developed in Schadt & al, 1998.

        Normalized such that the substitution rate on a stationnary
        sequence is 1.

        The only hypothesis is that the transversion rates depend only
        on the target nucleotides and not on the changed nucleotides.

        alpha: A->G
        beta: C->T
        gamma: A->C=G->C
        delta: C->A=T->A
        epsilon: G->A
        kappa: C->G=T->G
        lambda: A->T=G->T
        sigma: T->C
        
        eta: rate (default=1)
        """
        alpha=kw.get("alpha",1)
        beta=kw.get("beta",1)
        gamma=kw.get("gamma",1)
        delta=kw.get("delta",1)
        epsilon=kw.get("epsilon",1)
        kappa=kw.get("kappa",1)
        l=kw.get("lambda",1)
        sigma=kw.get("sigma",1)
        eta=kw.get("eta",1)

        for s in ["alpha","beta", "gamma", "delta", "epsilon", "kappa", "l", "sigma", "eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        piA=float(epsilon*(delta+kappa)+delta*(gamma+l))/((alpha+gamma+epsilon+l)*(gamma+delta+kappa+l))
        piG=float(alpha*(delta+kappa)+kappa*(gamma+l))/((alpha+gamma+epsilon+l)*(gamma+delta+kappa+l))
        piC=float(gamma*(delta+kappa)+sigma*(gamma+l))/((beta+delta+kappa+sigma)*(gamma+delta+kappa+l))
        piT=float(l*(delta+kappa)+beta*(gamma+l))/((beta+delta+kappa+sigma)*(gamma+delta+kappa+l))

        r=piA*(alpha+gamma+l)+piG*(epsilon+gamma+l)+piC*(delta+kappa+beta)+piT*(delta+kappa+sigma)

        eta2=eta/r
            
        s=""
        s+="A|C %f\n"%(eta2*gamma)
        s+="A|G %f\n"%(eta2*alpha)
        s+="A|T %f\n"%(eta2*l)
        s+="C|A %f\n"%(eta2*delta)
        s+="C|G %f\n"%(eta2*kappa)
        s+="C|T %f\n"%(eta2*beta)
        s+="G|A %f\n"%(eta2*epsilon)
        s+="G|C %f\n"%(eta2*gamma)
        s+="G|T %f\n"%(eta2*l)
        s+="T|A %f\n"%(eta2*delta)
        s+="T|C %f\n"%(eta2*sigma)
        s+="T|G %f\n"%(eta2*kappa)
    
        return s

    @staticmethod
    def RN95s(**kw):
        """ intersection between L95 and RN95: The two hypotheses are
        that the transversion rates are only dependent of the target
        nucleotide, and strand symmetry. 

        Normalized such that the substitution rate on a stationnary
        sequence is 1.

        alpha: A->G=T->C
        beta: C->T=G->A
        gamma: A->C=G->C=T->G=C->G
        delta: C->A=T->A=A->T=G->T
        
        eta: rate (default=1)
        """
        alpha=kw.get("alpha",1)
        beta=kw.get("beta",1)
        gamma=kw.get("gamma",1)
        delta=kw.get("delta",1)
        eta=kw.get("eta",1)

        for s in ["alpha","beta", "gamma", "delta","eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        piA=float((beta+delta)*(delta+gamma))/((alpha+gamma+beta+delta)*2*(gamma+delta))
        piG=float((alpha+gamma)*(delta+gamma))/((alpha+gamma+beta+delta)*2*(gamma+delta))
        piC=piG
        piT=piA

        r=piA*(alpha+gamma+delta)+piG*(beta+gamma+delta)+piC*(delta+gamma+beta)+piT*(delta+gamma+alpha)

        eta2=eta/r
    
        s=""
        s+="A|C %f\n"%(eta2*gamma)
        s+="A|G %f\n"%(eta2*alpha)
        s+="A|T %f\n"%(eta2*delta)
        s+="C|A %f\n"%(eta2*delta)
        s+="C|G %f\n"%(eta2*gamma)
        s+="C|T %f\n"%(eta2*beta)
        s+="G|A %f\n"%(eta2*beta)
        s+="G|C %f\n"%(eta2*gamma)
        s+="G|T %f\n"%(eta2*delta)
        s+="T|A %f\n"%(eta2*delta)
        s+="T|C %f\n"%(eta2*alpha)
        s+="T|G %f\n"%(eta2*gamma)
    
        return s

    @staticmethod
    def T92(**kw):
        """ Tamura 1992.
        Normalised such that the substitution rate on a stationnary
        sequence is 1.
        
        kappa: transition/transversion rate (default: 1).
        theta: equilibrium frequency of GC (default: 0.5).
        eta: rate (default=1)
        """
        kappa=kw.get("kappa",1)
        theta=kw.get("theta",0.5)
        eta=kw.get("eta",1)
    
        for s in ["kappa","theta", "eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)
        
        for s in ["theta"]:
            exec("if "+s+" >1: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        eta2=eta/(1.0+2*theta*kappa-2*theta*theta*kappa)

        s=""
        s+="A|C %f\n"%(theta*eta2)
        s+="A|G %f\n"%(theta*kappa*eta2)
        s+="A|T %f\n"%((1-theta)*eta2)
        s+="C|A %f\n"%((1-theta)*eta2)
        s+="C|G %f\n"%(theta*eta2)
        s+="C|T %f\n"%((1-theta)*kappa*eta2)
        s+="G|A %f\n"%((1-theta)*kappa*eta2)
        s+="G|C %f\n"%(theta*eta2)
        s+="G|T %f\n"%((1-theta)*eta2)
        s+="T|A %f\n"%((1-theta)*eta2)
        s+="T|C %f\n"%(theta*kappa*eta2)
        s+="T|G %f\n"%(theta*eta2)
        
        return s

    @staticmethod
    def HKY85(**kw):
        """ Hasegawa Kishino Yano 1985
        Normalised such that the substitution rate on a stationnary
        sequence is 1.

        kappa: transition/transversion rate
        theta: equilibrium frequency of GC.
        theta1: ratio of equilibrium frequency of A on equilibrium
                frequency of AT
        theta2: ratio of equilibrium frequency of G on equilibrium
                frequency of CG

         eta: rate (default=1)
         """
        
        kappa=kw.get("kappa",1)
        theta=kw.get("theta",0.5)
        theta1=kw.get("theta1",0.5)
        theta2=kw.get("theta2",0.5)
        eta=kw.get("eta",1)
    
        for s in ["kappa","theta","theta1","theta2","eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"'")

        for s in ["theta","theta1","theta2"]:
            exec("if "+s+" >1: raise ValueError, 'Bad value for "+s+"'")

        piA=theta1*(1-theta)
        piC=(1-theta2)*theta
        piG=theta2*theta
        piT=(1-theta1)*(1-theta)
        
        eta2=eta/(2.0*(piA*piC+piC*piG+piA*piT+piG*piT+kappa*(piC*piT+piA*piG)))

        s=""
        s+="A|C %f\n"%(piC*eta2)
        s+="A|G %f\n"%(kappa*piG*eta2)
        s+="A|T %f\n"%(piT*eta2)
        s+="C|A %f\n"%(piA*eta2)
        s+="C|G %f\n"%(piG*eta2)
        s+="C|T %f\n"%(kappa*piT*eta2)
        s+="G|A %f\n"%(kappa*piA*eta2)
        s+="G|C %f\n"%(piC*eta2)
        s+="G|T %f\n"%(piT*eta2)
        s+="T|A %f\n"%(piA*eta2)
        s+="T|C %f\n"%(kappa*piC*eta2)
        s+="T|G %f\n"%(piG*eta2)

        return s

    @staticmethod
    def TN93(**kw):
        """ Tamura and Nei 1993.
        Normalised such that the substitution rate on a stationnary
        sequence is 1.

        kappa1: transition/transversion for A<->G.
        kappa2: transition/transversion for C<->T.
        theta: equilibrium frequency of GC.
        theta1: ratio of equilibrium frequency of A on equilibrium
                frequency of AT
        theta2: ratio of equilibrium frequency of G on equilibrium
                frequency of CG
    
        eta: rate (default=1)
        """

        kappa1=kw.get("kappa1",1)
        kappa2=kw.get("kappa2",1)
        theta=kw.get("theta",0.5)
        theta1=kw.get("theta1",0.5)
        theta2=kw.get("theta2",0.5)
        eta=kw.get("eta",1)

        for s in ["kappa1","kappa2","theta","theta1","theta2","eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        for s in ["theta","theta1","theta2"]:
            exec("if "+s+" >1: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        piA=theta1*(1-theta)
        piC=(1-theta2)*theta
        piG=theta2*theta
        piT=(1-theta1)*(1-theta)

        eta2=eta/(2.0*(piA*piC+piC*piG+piA*piT+piG*piT+kappa2*piC*piT+kappa1*piA*piG))

        s=""
        s+="A|C %f\n"%(piC*eta2)
        s+="A|G %f\n"%(kappa1*piG*eta2)
        s+="A|T %f\n"%(piT*eta2)
        s+="C|A %f\n"%(piA*eta2)
        s+="C|G %f\n"%(piG*eta2)
        s+="C|T %f\n"%(kappa2*piT*eta2)
        s+="G|A %f\n"%(kappa1*piA*eta2)
        s+="G|C %f\n"%(piC*eta2)
        s+="G|T %f\n"%(piT*eta2)
        s+="T|A %f\n"%(piA*eta2)
        s+="T|C %f\n"%(kappa2*piC*eta2)
        s+="T|G %f\n"%(piG*eta2)

        return s

    @staticmethod
    def F84(**kw):    
        """ Felsenstein 1984.
        Normalised such that the substitution rate on a stationnary
        sequence is 1.

        kappa:  ratio transition/transversion
        theta:  equilibrium frequency of GC.
        theta1: ratio of equilibrium frequency of A on equilibrium
            frequency of AT
        theta2: ratio of equilibrium frequency of G on equilibrium
            frequency of CG
    
        eta: rate (default=1)
        """
    
        kappa=kw.get("kappa",1)
        theta=kw.get("theta",0.5)
        theta1=kw.get("theta1",0.5)
        theta2=kw.get("theta2",0.5)
        eta=kw.get("eta",1)
        
        for s in ["kappa","theta","theta1","theta2","eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        for s in ["theta","theta1","theta2"]:
            exec("if "+s+" >1: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        piA=theta1*(1-theta)
        piC=(1-theta2)*theta
        piG=theta2*theta
        piT=(1-theta1)*(1-theta)
        piR=piA+piG
        piY=piC+piT
    
        eta2=eta/(2.0*kappa*(piC*piT/(piC+piT)+piA*piG/(piA+piG))-piC*piC-piG*piG-piT*piT-piA*piA+1)

        s=""
        s+="A|C %f\n"%(piC*eta2)
        s+="A|G %f\n"%((1+kappa/piR)*piG*eta2)
        s+="A|T %f\n"%(piT*eta2)
        s+="C|A %f\n"%(piA*eta2)
        s+="C|G %f\n"%(piG*eta2)
        s+="C|T %f\n"%((1+kappa/piY)*piT*eta2)
        s+="G|A %f\n"%((1+kappa/piR)*piA*eta2)
        s+="G|C %f\n"%(piC*eta2)
        s+="G|T %f\n"%(piT*eta2)
        s+="T|A %f\n"%(piA*eta2)
        s+="T|C %f\n"%((1+kappa/piY)*piC*eta2)
        s+="T|G %f\n"%(piG*eta2)

        return s

    @staticmethod
    def GTR(**kw):
        """ General Time-Reversible substitution model,
        parametrization of Yang 1994.
        Normalised such that the substitution rate on a stationnary
        sequence is 1.

        In proportion to the A<->G rate, the arguments are (default 1):

        a: C<->T
        b: A<->T
        c: G<->T
        d: A<->C
        e: C<->G

        theta: equilibrium frequency of GC.
        theta1: ratio of equilibrium frequency of A on equilibrium
                frequency of AT
        theta2: ratio of equilibrium frequency of G on equilibrium
                frequency of CG
        eta: rate (default=1)
        """

        a=kw.get("a",1)
        b=kw.get("b",1)
        c=kw.get("c",1)
        d=kw.get("d",1)
        e=kw.get("e",1)
        theta=kw.get("theta",0.5)
        theta1=kw.get("theta1",0.5)
        theta2=kw.get("theta2",0.5)
        eta=kw.get("eta",1)

        for s in ["a", "b", "c", "d", "e", "theta","theta1","theta2", "eta"]:
            exec("if "+s+" <0: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        for s in ["theta","theta1","theta2"]:
            exec("if "+s+" >1: raise ValueError, 'Bad value for "+s+"(%f)'%"+s)

        piA=theta1*(1-theta)
        piC=(1-theta2)*theta
        piG=theta2*theta
        piT=(1-theta1)*(1-theta)
        piR=piA+piG
        piY=piC+piT

        eta2=eta/(2.0*(a*piC*piT+b*piA*piT+c*piC*piG+d*piA*piC+e*piC*piG+piA*piG))
    
        s=""
        s+="A|C %f\n"%(d*piC*eta2)
        s+="A|G %f\n"%(piG*eta2)
        s+="A|T %f\n"%(b*piT*eta2)
        s+="C|A %f\n"%(d*piA*eta2)
        s+="C|G %f\n"%(e*piG*eta2)
        s+="C|T %f\n"%(a*piT*eta2)
        s+="G|A %f\n"%(piA*eta2)
        s+="G|C %f\n"%(e*piC*eta2)
        s+="G|T %f\n"%(c*piT*eta2)
        s+="T|A %f\n"%(b*piA*eta2)
        s+="T|C %f\n"%(a*piC*eta2)
        s+="T|G %f\n"%(c*piG*eta2)

        return s

    
class NuclModel(Model):
    """ Class of Models defined from usual single-site nucleotide
models.

        Context-dependent substitution rates can be ADDED to the
Model. They are denoted as ruXvY, which describes the substitution
from base X with left context word u and right context word v to base
Y. The rate is a factor of X->Y substitution rate in the context
independent model.

        The syntax to build a model is like:
        
        m=NuclModel(model="T92(theta=0.1,rCgT=10)")
"""
    
    def __init__(self,**kw):
        if not kw.has_key("model"):
            Model.__init__(self)
            return

        func=kw["model"]
        Model.__init__(self,str=eval("BibliNucl."+func))

        def loc(**kw):
            return locals()
        pp=func.find("(")
        dl=eval("loc"+func[pp:])["kw"]

        for k in dl:
            if k[0]=="r":
                i=1
                while i<len(k):
                    if k[i].isupper():
                        break
                    i+=1
                else:
                    continue

                if k[-1].islower():
                    continue

                ex=0
                dne=self.next_extended(k[i])
                for [a,b,c,d] in dne:
                    if c==k[-1] and a=="" and b=="":
                        ex=d
                        break

                self._Proportion__add(k[1:-1],k[-1],ex*dl[k])

        self.update()
